using System;
using System.Collections.Generic;
using LinearAlgebra;

namespace RoboidControl {

    /// <summary>
    /// A thing is the primitive building block
    /// </summary>
    [Serializable]
    public class Thing {

        #region Types

        /// <summary>
        /// Predefined thing types
        /// </summary>
        public enum Type {
            Undetermined,
            // Sensor
            Switch,
            DistanceSensor,
            DirectionalSensor,
            TemperatureSensor,
            TouchSensor,
            // Motor
            ControlledMotor,
            UncontrolledMotor,
            Servo,
            // Other
            Roboid,
            Humanoid,
            ExternalSensor
        };

        public delegate void ChangeHandler();
        public delegate void SphericalHandler(Spherical v);
        public delegate void ThingHandler(Thing t);

        #endregion Types

        #region Init

        /// <summary>
        /// Create a new thing for a participant
        /// </summary>
        /// <param name="owner">The participant owning the thing</param>
        /// <param name="thingType">The type of thing</param>
        public Thing(Participant owner, byte thingType = (byte)Type.Undetermined, bool invokeEvent = true) {
            this.owner = owner;
            this.type = thingType;
            if (this.owner != null)
                this.owner.Add(this);
            if (invokeEvent)
                InvokeNewThing(this);
        }
        public Thing(Participant owner) : this(owner, Type.Undetermined) { }
        /// <summary>
        /// Create a new thing for a participant
        /// </summary>
        /// <param name="owner">The participant owning the thing</param>
        /// <param name="thingType">The type of thing</param>
        public Thing(Participant owner, Type thingType = Type.Undetermined, bool invokeEvent = true) : this(owner, (byte)thingType, invokeEvent) {
        }
        /// <summary>
        /// Create a new thing without communication abilities
        /// </summary>
        /// <param name="thingType">The type of thing</param>
        public Thing(byte thingType = (byte)Type.Undetermined, bool invokeEvent = true) : this(ParticipantUDP.Isolated(), thingType, invokeEvent) {
        }

        /// <summary>
        /// Create a new thing as a child of another thing
        /// </summary>
        /// <param name="parent">The parent thing</param>
        /// <param name="thingType">The type of thing</param>
        public Thing(Thing parent, byte thingType = (byte)Type.Undetermined, bool invokeEvent = true) : this(parent.owner, thingType, invokeEvent) {
            this.parent = parent;
        }

        /// <summary>
        /// Create a new thing for the given participant
        /// </summary>
        /// <param name="owner">The participant owning the thing</param>
        /// <param name="networkId">The network ID of the thing</param>
        /// <param name="thingId">The ID of the thing</param>
        /// <param name="thingType">The type of thing</param>
        public Thing(Participant owner, byte networkId, byte thingId, byte thingType = 0) {
            this.owner = owner;
            this.id = thingId;
            this.type = thingType;
            this.networkId = networkId;
            // Console.Write($"New thing added to {owner}");
            this.owner.Add(this);
            InvokeNewThing(this);
        }

        /// <summary>
        /// Function which can be used to create components in external engines.
        /// </summary>
        /// Currently this is used to create GameObjects in Unity
        public virtual void CreateComponent() {
#if UNITY_5_3_OR_NEWER            
            this.component = Unity.Thing.Create(this);
            this.component.core = this;
#endif            
        }

        #endregion Init

        #region Properties

        /// <summary>
        /// The participant owning this thing
        /// </summary>
        public Participant owner = null;
        /// <summary>
        /// The network ID of this thing.
        /// </summary>
        /// @note This field will likely disappear in future versions
        public byte networkId = 0;
        /// <summary>
        /// The ID of this thing
        /// </summary>
        public byte id = 0;

        /// <summary>
        /// The type of this thing. 
        /// </summary>
        /// This can be either a Thing::Type (needs casting) or a byte value for custom types.
        public byte type = (byte)Type.Undetermined;

        private Thing _parent;
        /// <summary>
        /// The parent of this thing
        /// </summary>
        public Thing parent {
            get => _parent;
            set {
                if (_parent == value)
                    return;

                if (value == null) {
                    _parent?.RemoveChild(this);
                    _parent = null;
                }
                else {
                    value.AddChild(this);
                }
                this.hierarchyChanged = true;
            }
        }

        /// <summary>
        /// Add a child Thing to this Thing
        /// </summary>
        /// <param name="child">The Thing which should become a child</param>
        /// @remark When the Thing is already a child, it will not be added again
        public void AddChild(Thing child) {
            if (children.Find(thing => thing == child) != null)
                return;

            child._parent = this;
            children.Add(child);
        }
        /// <summary>
        /// Remove the given thing as a child of this thing
        /// </summary>
        /// <param name="child">The child to remove</param>
        /// <returns>True when the child was present or false when it was not found</returns>
        public bool RemoveChild(Thing child) {
            return children.Remove(child);
        }

        /// <summary>
        /// Get a child by thing Id
        /// </summary>
        /// <param name="thingId"></param>
        /// <param name="recursively"></param>
        /// <returns></returns>
        Thing GetChild(byte thingId, bool recursively = false) {
            foreach (Thing child in this.children) {
                if (child == null)
                    continue;

                if (child.id == thingId)
                    return child;
                if (recursively) {
                    Thing foundChild = child.GetChild(thingId, recursively);
                    if (foundChild != null)
                        return foundChild;
                }
            }
            return null;
        }
        /// <summary>
        /// Find a child by name
        /// </summary>
        /// <param name="name">The name of the child thing</param>
        /// <param name="recursively">If true, the name will be searched through descendants recursively</param>
        /// <returns>The found thing or null when nothing is found</returns>
        Thing FindChild(string name, bool recursively = true) {
            foreach (Thing child in this.children) {
                if (child == null)
                    continue;

                if (child.name == name)
                    return child;

                if (recursively) {
                    Thing foundChild = child.FindChild(name, recursively);
                    if (foundChild != null)
                        return foundChild;
                }
            }
            return null;
        }

        /// <summary>
        /// Indicator that the hierarchy of the thing has changed
        /// </summary>
        public bool hierarchyChanged = true;

        /// <summary>
        /// The children of this thing
        /// </summary>
        [NonSerialized]
        protected List<Thing> children = new();

        private string _name = "";
        /// <summary>
        /// The name of the thing
        /// </summary>
        public virtual string name {
            get => _name;
            set {
                if (_name != value) {
                    _name = value;
                    nameChanged = true;
                    OnNameChanged?.Invoke();
                }
            }
        }
        /// <summary>
        /// Event which is triggered when the name changes
        /// </summary>
        public event ChangeHandler OnNameChanged = delegate { };
        public bool nameChanged = false;

        /// <summary>
        /// An URL pointing to the location where a model of the thing can be found
        /// </summary>
        public string modelUrl = "";

        private Spherical _position = Spherical.zero;
        /// <summary>
        /// The position of the thing in local space, in meters.
        /// </summary>
        public Spherical position {
            get { return _position; }
            set {
                if (_position != value) {
                    _position = value;
                    positionUpdated = true;
                    //OnPositionChanged?.Invoke();
                }
            }
        }
        /// <summary>
        /// Event triggered when the pose has changed
        /// </summary>
        public event ChangeHandler OnPoseChanged = null; //delegate { };
        /// <summary>
        /// Boolean indicating that the thing has an updated position
        /// </summary>
        public bool positionUpdated = false;

        private SwingTwist _orientation = SwingTwist.zero;
        /// <summary>
        /// The orientation of the thing in local space
        /// </summary>
        public SwingTwist orientation {
            get { return _orientation; }
            set {
                if (_orientation != value) {
                    _orientation = value;
                    orientationUpdated = true;
                    //OnOrientationChanged?.Invoke();
                }
            }
        }
        /// <summary>
        /// Boolean indicating the thing has an updated orientation
        /// </summary>
        public bool orientationUpdated = false;

        private Spherical _linearVelocity = Spherical.zero;
        /// <summary>
        /// The linear velocity of the thing in local space in meters per second
        /// </summary>
        public Spherical linearVelocity {
            get => _linearVelocity;
            set {
                if (_linearVelocity != value) {
                    _linearVelocity = value;
                    linearVelocityUpdated = true;
                    OnLinearVelocityChanged?.Invoke(_linearVelocity);
                }
            }
        }
        /// <summary>
        /// Event triggered when the linear velocity has changed
        /// </summary>
        public event SphericalHandler OnLinearVelocityChanged = delegate { };
        /// <summary>
        /// Boolean indicating the thing has an updated linear velocity
        /// </summary>
        public bool linearVelocityUpdated = false;

        private Spherical _angularVelocity = Spherical.zero;
        /// <summary>
        /// The angular velocity of the thing in local space in degrees per second
        /// </summary>
        public Spherical angularVelocity {
            get => _angularVelocity;
            set {
                if (_angularVelocity != value) {
                    _angularVelocity = value;
                    angularVelocityUpdated = true;
                    OnAngularVelocityChanged?.Invoke(_angularVelocity);
                }
            }
        }
        /// <summary>
        /// Event triggered when the angular velocity has changed
        /// </summary>
        public event SphericalHandler OnAngularVelocityChanged = delegate { };
        /// <summary>
        /// Boolean indicating the thing has an updated angular velocity
        /// </summary>
        public bool angularVelocityUpdated = false;

#if UNITY_5_3_OR_NEWER
        /// <summary>
        /// A reference to the representation of the thing in Unity
        /// </summary>
        [NonSerialized]
        public Unity.Thing component = null;
#endif

        public bool terminate = false;

        #endregion Properties

        #region Methods

        public static ulong GetTimeMs() {
#if UNITY_5_3_OR_NEWER
            return (ulong)(UnityEngine.Time.time * 1000);
#endif
        }

        /// <summary>
        /// Update de state of the thing
        /// </summary>
        /// <param name="recursively">When true, this will Update the descendants recursively</param>
        public void Update(bool recursively = false) {
            Update(TimeManager.GetCurrentTimeMilliseconds(), recursively);
        }
        // #endif
        /// <summary>
        /// Update this thing
        /// </summary>
        /// <param name="currentTime">The current time in milliseconds</param>
        public virtual void Update(ulong currentTimeMs, bool recursively = false) {
            if (this.positionUpdated || this.orientationUpdated)
                OnPoseChanged?.Invoke();
            this.positionUpdated = false;
            this.orientationUpdated = false;
            this.linearVelocityUpdated = false;
            this.angularVelocityUpdated = false;
            //this.hierarchyChanged = false;

            // should recurse over children...
            if (recursively) {
                for (byte childIx = 0; childIx < this.children.Count; childIx++) {
                    Thing child = this.children[childIx];
                    if (child == null)
                        continue;
                    child.Update(currentTimeMs, recursively);
                }
            }
        }

        /// <summary>
        /// Function used to generate binary data for this thing
        /// </summary>
        /// <returns>a byte array with the binary data</returns>
        /// @sa Passer::RoboidControl::BinaryMsg
        public virtual byte[] GenerateBinary() { return Array.Empty<byte>(); }

        /// <summary>
        /// Function used to process binary data received for this thing
        /// </summary>
        /// <param name="bytes">The binary data</param>
        public virtual void ProcessBinary(byte[] bytes) {
        }

        #endregion Methods

        /// <summary>
        /// Event triggered when a new thing has been created
        /// </summary>
        public static event ThingHandler OnNewThing = delegate { };
        /// <summary>
        /// Trigger the creation for the given thing
        /// </summary>
        /// <param name="thing">The created thing</param>
        public static void InvokeNewThing(Thing thing) {
            OnNewThing?.Invoke(thing);
        }

        /// <summary>
        /// Check if the thing has the given properaties
        /// </summary>
        /// <param name="thing">The thing to check</param>
        /// <param name="networkId">The network ID to compare to</param>
        /// <param name="thingId">The thing ID to compare to</param>
        /// <returns>True when the thing has the given properties</returns>
        public static bool IsThing(Thing thing, byte networkId, byte thingId) {
            if (thing == null)
                return false;
            return (thing.networkId == networkId) && (thing.id == thingId);
            //return (thing.id == thingId);
        }

    }
}
